"""Parameter file is read into memory and modified there"""

import string,sys

from FileBasis import FileBasisBackup
from PyFoam.Basics.PlyParser import PlyParser
from PyFoam.Basics.FoamFileGenerator import FoamFileGenerator

class ParsedParameterFile(FileBasisBackup):
    """ Parameterfile whose complete representation is read into
    memory, can be manipulated and afterwards written to disk"""

    def __init__(self,name,backup=False,debug=False,boundaryDict=False):
        """@param name: The name of the parameter file
        @param backup: create a backup-copy of the file
        @param boundaryDict: the file to parse is a boundary file"""

        FileBasisBackup.__init__(self,name,backup=backup)
        self.debug=debug
        self.boundaryDict=boundaryDict
        
        self.header=None
        self.content=None

        self.readFile()

    def parse(self,content):
        """Constructs a representation of the file"""
        parser=FoamFileParser(content,debug=self.debug,boundaryDict=self.boundaryDict)
        self.content=parser.getData()
        self.header=parser.getHeader()
        return self.content

    def __getitem__(self,key):
        return self.content[key]

    def __setitem__(self,key,value):
        self.content[key]=value

    def makeString(self):
        """Generates a string from the contents in memory"""
        string="// File generated by PyFoam - sorry for the ugliness\n\n"
##        if self.header:
##            generator=FoamFileGenerator(self.header)
##            string+=generator.makeString()
##            string+="\n// End of header\n\n"

        generator=FoamFileGenerator(self.content,header=self.header)
        string+=generator.makeString()

        return string

class FoamFileParser(PlyParser):
    """Class that parses a string that contains the contents of an
    OpenFOAM-file and builds a nested structure of directories and
    lists from it"""

    def __init__(self,content,debug=False,noHeader=False,boundaryDict=False):
        """@param content: the string to be parsed
        @param debug: output debug information during parsing
        @param noHeader: switch that turns off the parsing of the header"""

        self.data=None
        self.header=None
        self.debug=debug

        if noHeader:
            self.start='noHeader'

        if boundaryDict:
            self.start='boundaryDict'

        PlyParser.__init__(self,debug=debug)

        #sys.setrecursionlimit(50000)
        #print sys.getrecursionlimit()
        
        self.header,self.data=self.parse(content)

    def getData(self):
        """ Get the data structure"""
        return self.data

    def getHeader(self):
        """ Get the OpenFOAM-header"""
        return self.header

    def printContext(self,c,ind):
        """Prints the context of the current index"""
        print "------"
        print c[max(0,ind-100):max(0,ind-1)]
        print "------"
        print ">",c[ind-1],"<"
        print "------"
        print c[min(len(c),ind):min(len(c),ind+100)]
        print "------"

    def parserError(self,text,c,ind):
        """Prints the error message of the parser and exit"""
        print "PARSER ERROR:",text
        print "On index",ind
        self.printContext(c,ind)
        raise ParseError
        sys.exit(-1)

    tokens = (
        'NAME',
        'ICONST',
        'FCONST',
        'SCONST',
        'FOAMFILE',
        'UNIFORM',
        'NONUNIFORM',
    )

    reserved = {
        'FoamFile'   : 'FOAMFILE',
        'uniform'    : 'UNIFORM',
        'nonuniform' : 'NONUNIFORM',
        }

    def t_NAME(self,t):
        r'[a-zA-Z_][<>(),.\*|a-zA-Z_0-9]*'
        t.type=self.reserved.get(t.value,'NAME')
        return t

    t_ICONST = r'(-|)\d+([uU]|[lL]|[uU][lL]|[lL][uU])?'

    t_FCONST = r'(-|)((\d+)(\.\d+)(e(\+|-)?(\d+))? | (\d+)e(\+|-)?(\d+))([lL]|[fF])?'

    t_SCONST = r'\"([^\\\n]|(\\.))*?\"'

    literals = "(){};[]"

    t_ignore=" \t"

    # Define a rule so we can track line numbers
    def t_newline(self,t):
        r'\n+'
        t.lexer.lineno += len(t.value)

    # C or C++ comment (ignore)
    def t_ccode_comment(self,t):
        r'(/\*(.|\n)*\*/)|(//.*)'
        t.lexer.lineno += t.value.count('\n')
        pass

    # Error handling rule
    def t_error(self,t):
        print "Illegal character '%s'" % t.value[0]
        t.lexer.skip(1)

    def p_global(self,p):
        'global : header dictbody'
        p[0] = ( p[1] , p[2] )

    def p_noHeader(self,p):
        'noHeader : dictbody'
        p[0] = ( None , p[1] )

    def p_boundaryDict(self,p):
        'boundaryDict : header list'
        #        p[0] = (  p[1] , dict(zip(p[2][::2],p[2][1::2])) )
        p[0] = (  p[1] , p[2] )

    def p_header(self,p):
        'header : FOAMFILE dictionary'
        p[0] = p[2]

    def p_dictionary(self,p):
        '''dictionary : '{' dictbody '}'
                      | '{' '}' '''
        if len(p)==4:
            p[0] = p[2]
        else:
            p[0] = {}

    def p_dictbody(self,p):
        '''dictbody : dictbody dictline
                    | dictline
                    | empty'''

        if len(p)==3:
            p[0]=p[1]
            p[0][p[2][0]]=p[2][1]
        else:
            if p[1]:
                p[0]={p[1][0]:p[1][1]}
            else:
                p[0]={}

    def p_list_normal(self,p):
        '''list : '(' itemlist ')' '''
        p[0] = p[2]

    def p_list_prefixed(self,p):
        '''list : ICONST '(' itemlist ')' '''
        p[0] = p[3]

    def p_itemlist(self,p):
        '''itemlist : itemlist item
                    | item '''
        if len(p)==2:
            if p[1]==None:
                p[0]=[]
            else:
                p[0]=[ p[1] ]
        else:
            p[0]=p[1]
            p[0].append(p[2])

    def p_dictline(self,p):
        '''dictline : NAME dictitem ';'
                    | NAME list ';'
                    | NAME fieldvalue ';'
                    | NAME dictionary'''
        p[0]= ( p[1] , p[2] )

    def p_number(self,p):
        '''number : ICONST
                  | FCONST'''
        p[0] = p[1]

    def p_dimension(self,p):
        '''dimension : '[' number number number number number number number ']' '''
        p[0]=string.join(p[1:])

    def p_vector(self,p):
        '''vector : '(' number number number ')' '''
        p[0]=string.join(p[1:])

    def p_fieldvalue_uniform(self,p):
        '''fieldvalue : UNIFORM number
                      | UNIFORM vector'''
        p[0] = p[1]+" "+p[2]

    def p_fieldvalue_nonuniform(self,p):
        '''fieldvalue : NONUNIFORM NAME list'''
        p[0] = (p[1]+" "+p[2] , p[3])

    def p_dictitem(self,p):
        '''dictitem : longitem
                    | pitem'''
        p[0] = p[1]

    def p_longitem(self,p):
        '''longitem : pitemlist pitem'''
        p[0] = p[1]+(p[2],)

    def p_pitemlist(self,p):
        '''pitemlist : pitemlist pitem
                     | pitem '''
        if len(p)==2:
            p[0]=(p[1],)
        else:
            p[0]=p[1]+(p[2],)

    def p_pitem(self,p):
        '''pitem : NAME
                 | SCONST
                 | number
                 | dimension
                 | empty'''
        p[0] = p[1]

    def p_item(self,p):
        '''item : pitem
                | list
                | dictionary'''
        p[0] = p[1]

    def p_empty(self,p):
        'empty :'
        pass

    def p_error(self,p):
        print "Syntax error at token", p # .type, p.lineno
        # Just discard the token and tell the parser it's okay.
        self.yacc.errok()

